#!/usr/bin/perl -w

##############################################################################
#
# oracle_osaudittrail_fwdr - 
#	reads ddl and dml logs and formats the logs and forwards via tcp/udp
#
# original author: Rory Bray
# leech author: Amer Hassounah
# date: November 24, 2009
#
# Copyright Q1 Labs, Inc. 2009
# Revision 1.0
##############################################################################

use strict;
use Getopt::Std;
use File::Basename;
use Fcntl qw(LOCK_EX LOCK_NB);
use Socket;
use File::Monitor;

my %options = (
	D => 0,
	n => 0,
	d => "",
	f => "user.info",
	t => "localhost",
	p => 514,
	H => `hostname`,
	r => "/var/run",
	l => "/var/lock",
	u => 0,
	h => 0,
	v => 0,
	V => 0
);
getopts( 'Dnuht:d:f:p:H:r:l:vV', \%options );

sub usage()
{
	print <<"DONE";

Usage: $0 -t target_host -d logs_directory [options]

	-h	Display this help message and exit
	
	-H	The hostname which will be used in the syslog header

	-t	The target remote host to which messages will be sent

	-D 	Do not daemon-ize. by default the script will run in the background,
		redirect stdout and stderr to syslog and create a pidfile. -D prevents all this
		and the script will run in the foreground.

	-n 	Process new logs only, existing logs will be watched and new events added will be processed.

	-d 	The directory where the DDL and DML logs are located.

	-u 	Use udp instead of tcp. the default is tcp for reliable delivery and because
		log messages will be truncated via udp. udp will perform better, however.

	-f 	Syslog facility which will be used in the syslog header. When
		the header is added the default facility.severity will be user.info. use -f to
		override this if needed (e.g. local0.notice or daemon.info)

	-r 	The directory location for the .pid file (not used if -D specified)

	-l 	The directory location for the .lock file (not used if -D specified)

	-v	Prints the version of the script 

	-V	Verbose Debug Logging Enabled
DONE
}

if($options{h})
{
    usage();
	exit(1);
}

if($options{v})
{
	print "Script Version -- 5.3\n";
	exit(1);
}

# If no options have been specified, print usage and exit.
if($options{d} eq "")
{
        print "No directory specified for monitoring, please specify one!\n";
        usage();
        exit(1);
}

# cleanup/validate options
$options{H} =~ m/^(.*)$/;
my $syslogHost = $1;
my $dirPath;
if ($options{d} =~ m/.*\/$/)
{
	$dirPath = $options{d};
} else
{
	$dirPath = $options{d}."/";
}

# the following are used for process control
my $execName = basename($0);
my $lockFile = $options{l} . "/" . $execName . ".lock";
my $pidFile  = $options{r} . "/" . $execName . ".pid";

# loop/exit condition
my $running  = 1;
my $continue = 0;

my $lastMessage = 0;

# ugly old month-map for formatting syslog headers
my %monthMap = (
	"01" => "Jan",
	"02" => "Feb",
	"03" => "Mar",
	"04" => "Apr",
	"05" => "May",
	"06" => "Jun",
	"07" => "Jul",
	"08" => "Aug",
	"09" => "Sep",
	"10" => "Oct",
	"11" => "Nov",
	"12" => "Dec"
);

# even more ugly syslog facility/severity junk
my %syslog_facilities = (
	"kern"     => 0,
	"user"     => 8,
	"mail"     => 16,
	"daemon"   => 24,
	"auth"     => 32,
	"syslog"   => 40,
	"lpr"      => 48,
	"news"     => 56,
	"uucp"     => 64,
	"cron"     => 72,
	"authpriv" => 80,
	"ftp"      => 88,
	"local0"   => 128,
	"local1"   => 136,
	"local2"   => 144,
	"local3"   => 152,
	"local4"   => 160,
	"local5"   => 168,
	"local6"   => 176,
	"local7"   => 184
);

my %syslog_severities = (
	"emerg"   => 0,
	"alert"   => 1,
	"crit"    => 2,
	"err"     => 3,
	"warning" => 4,
	"notice"  => 5,
	"info"    => 6,
	"debug"   => 7
);

my $socket;    # global handles are evil but it's just easier to keep this guy in scope

sub syslogPri($) {
	( my $facSevString ) = @_;
	my $facility = "user";
	my $severity = "info";
	if ($facSevString =~ m/^(kern|user|mail|daemon|auth|syslog|lpr|news|uucp|cron|authpriv|ftp|local[0-7])/)
	{
		$facility = $1;
	}
	if (
		$facSevString =~ m/\.(emerg|alert|crit|err|warning|notice|info|debug)/ )
	{
		$severity = $1;
	}
	return $syslog_facilities{$facility} + $syslog_severities{$severity};
}

sub stubbornReconnect($$$) {
	( my $host, my $port, my $sleepTime ) = @_;
	while ( !$continue ) {
		disconnectMe($socket);    # just to be sure we're all cleaned up
		$socket = connectMe( $host, $port );
		if ( !$socket && $running ) {
			sleep $sleepTime;
			next;
		}
		$continue = 1;
	}
}

sub connectMe($$) {
	( my $host, my $port ) = @_;
	local *SOCK;
	my $proto    = ( getprotobyname('tcp') )[2];
	my $type	 = SOCK_STREAM;
	if ($options{u})
	{
		$proto    = getprotobyname('udp');
		$type	 = SOCK_DGRAM;
	} 
	my $ipAddr   = inet_aton($host);
	my $sockAddr = sockaddr_in( $port, $ipAddr );
	unless (socket( SOCK, PF_INET, $type, $proto ) )
	{
		print "unable to open socket: $!\n";
		return 0;
	}
	unless ( connect( SOCK, $sockAddr ) ) {
		print "unable to connect socket: $!\n";
		return 0;
	}
	select(SOCK);
	$| = 1;    # quick select to set default file handle and make it autoflush
	select(STDOUT);
	print "successfully connected to " . $host . ":" . $port . "\n";
	return *SOCK{IO};
}

sub disconnectMe($) {
	( my $socket ) = @_;
	if ($socket) {
		close($socket) or print "Unable to close socket: $!\n";
	}
	print "disconnecting\n";
}

sub startDaemon($$$) {
	( my $execName, my $lockFile, my $pidFile ) = @_;

# now, before loading config and all that, make sure we're the only daemon running
	open( SELFLOCK, ">$lockFile" ) or die("Couldn't open $lockFile: $!\n");
	flock( SELFLOCK, LOCK_EX | LOCK_NB )
	  or die("Aborting: another $execName is already running\n");

# Get ready to daemonize by redirecting our output to syslog, requesting that logger prefix the lines with our program name:
	open( STDOUT, "|-", "logger -t $execName" )
	  or die("Couldn't open logger output stream: $!\n");
	open( STDERR, ">&STDOUT" )
	  or die("Couldn't redirect STDERR to STDOUT: $!\n");
	$| = 1;   # Make output line-buffered so it will be flushed to syslog faster

	chdir('/')
	  ; # Avoid the possibility of our working directory resulting in keeping an otherwise unused filesystem in use

	# Double-fork to avoid leaving a zombie process behind:
	exit if ( fork() );
	exit if ( fork() );
	sleep 5;    # until getppid() == 1;

	# write the daemonized pid to pidfile for init script control
	open( PIDFILE, ">$pidFile" );
	print PIDFILE "$$\n";
	close(PIDFILE);

	print "$execName $$ successfully daemonized\n";
}

sub stopDaemon($$) {
	( my $lockFile, my $pidFile ) = @_;
	unlink $lockFile;
	unlink $pidFile;
}

my $currentRec = {
	TIMESTAMP   => "",
	ACTION      => "",
	DB_USER     => "",
	PRIVILEGE   => "",
	CL_USER     => "",
	CL_TERMINAL => "",
	STATUS      => "",
	SOURCE      => "",
};
my $monitor = File::Monitor->new();
my %allLogFiles;
my @unwatchFiles;

sub processFile($) {
	my $fileName  = shift;

	if (!exists $allLogFiles{$fileName})
	{
		#if no line marker exists assign line 0
		$allLogFiles{$fileName} = 0;
	}
	my $startLine = $allLogFiles{$fileName};
	
	#open the log file for reading
	unless ( open( LOG, $dirPath. $fileName ) ) {
		print "unable to launch log: $!\n";
		return 0;
	}
	print "logging started with " . $fileName . "\n";

	my $tempLine;

	while (<LOG>) {

		#keep looping through the input until we reach the desired line
		next unless $. > $startLine;

		$tempLine = $_;

#events start with a timestamp using this sample format: Thu Aug 20 12:11:00 2009
		if ( $tempLine =~
			m/^\w{3} \w{3} (?:\s\d|\d{2}) \d{2}:\d{2}:\d{2} \d{4}/ )
		{
			resetCurrentLogRecord();

			#$fileName =~ s/\r\n//;
			$currentRec->{SOURCE} = $fileName;
			$tempLine =~ s/\r\n//;
			chomp($tempLine);
			$currentRec->{TIMESTAMP} = $tempLine;
			while (<LOG>) {
				$tempLine = $_;
				$tempLine =~ s/\r\n//;
				if ( $tempLine =~ m/^DATABASE USER:(?:\[\d+\])? (.*)$/ ) {
					$currentRec->{DB_USER} = $1;
				}
				elsif ( $tempLine =~ m/^PRIVILEGE :(?:\[\d+\])? (.*)$/ ) {
					$currentRec->{PRIVILEGE} = $1;
				}
				elsif ( $tempLine =~ m/^CLIENT USER:(?:\[\d+\])? (.*)$/ ) {
					$currentRec->{CL_USER} = $1;
				}
				elsif ( $tempLine =~ m/^CLIENT TERMINAL:(?:\[\d+\])? (.*)$/ ) {
					$currentRec->{CL_TERMINAL} = $1;
				}
				elsif ( $tempLine =~ m/^STATUS:(?:\[\d+\])? (.*)$/ ) {
					$currentRec->{STATUS} = $1;
				}
				elsif ( $tempLine =~ m/^\n/ ) { last; }
				else {
					chomp($tempLine);
					$tempLine =~ s/^ACTION :(?:\[\d+\])? //;
					$currentRec->{ACTION} .= $tempLine;
				}
			}

			#capture the number of the last line processed
			$allLogFiles{$fileName} = $.;
			emitLog();
		}

	}
	close(LOG);
}

sub processDirectory($) {
	my $logDirectory = shift;
	my $file;
	print "Adding ".$logDirectory." to watch list\n";
	$monitor->watch(
		{
			name    => $logDirectory,
			recurse => 1,
			files   => 1,
			callback =>
			  sub {
				my ( $dirName, $event, $change ) = @_;
				#process any new files created if they match the file pattern
				my @newFiles = $change->files_created;
				my $newTempFile;
				foreach $newTempFile (@newFiles) {
					print "New File Name: " . $newTempFile ."\n";
					if ( $newTempFile =~ m/.*?([A-Za-z0-9-_]*?ora_\d+.*\.aud)$/ ) {
						print "New file detected, adding ".$newTempFile." to watch list\n";
						$allLogFiles{$1} = 0;
						$monitor->watch(
							{
								name => $newTempFile,
								callback => 
								  sub {
									my ( $changedFile, $event, $change ) = @_;
									$changedFile =~ m/.*?([A-Za-z0-9-_]*?ora_\d+.*\.aud)$/;

				                                        # We're seeing instances in file add/removal where change notifications are coming in for files that were previously deleted
				                                        # Setting it to only process here as a "change" if we've initialized the file in $allLogFiles. It will still get added later as "new file" otherwise
				                                        if (defined($change->new_size) && $change->new_size != 0 && defined($allLogFiles{$1}))
									{
										print "Process file via size change\n" if $options{V};
										processFile($1);	
									}
								  }
							}
						);
						print "Process file via add new\n" if $options{V};
						processFile($1);
					}
				}
				#process any files deleted if they match the file pattern
				my @deletedFiles = $change->files_deleted;
			  	my $deletedTempFile;
				foreach $deletedTempFile (@deletedFiles)
				{
					if ( $deletedTempFile =~ m/.*?([A-Za-z0-9-_]*?ora_\d+.*\.aud)$/ )
					{
						print "File ".$deletedTempFile." deleted: flagging for removal from watch list\n";
						push (@unwatchFiles,$1);
					}					
				}
			  }
		}
	);


	$monitor->scan;

	
	opendir( DIR, $logDirectory );
	my @files = grep( /^.*ora_\d+.*\.aud$/, readdir(DIR) );
	closedir(DIR);
	foreach $file (@files) {
		$allLogFiles{$file} = 0;
		my $filePath = $dirPath.$file;
		print "Adding ".$filePath." to watch list\n";
		$monitor->watch(
			{
				name => $filePath,
				callback => 
				  sub {
					my ( $changedFile, $event, $change ) = @_;
					$changedFile =~ m/.*?([A-Za-z0-9-_]*?ora_\d+.*\.aud)$/;
					if (defined($change->new_size) && $change->new_size != 0 && defined($allLogFiles{$1}))
					{
						print "Process file via directory add size change\n" if $options{V};
						processFile($1);	
					}
				  }
			}
		);
		$monitor->scan;
		if (!$options{n}){
			print "Process file via add new directory\n" if $options{V};
			processFile($file);
		} else 
		{
			my $lines = `wc -l $filePath`;
			$lines =~ m/^(\d+)/;
			$allLogFiles{$file} = $1;
		}
	}
}

sub emitLog() {
	my $msg;
	my $cmd;

	$lastMessage = 0;

	#debug
	#my $syslogPriTag = syslogPri("user.info");
	#my $syslogHost = "172.16.210.60";

	# Real
	my $syslogPriTag = syslogPri( $options{f} );

	# Create the header.
	(
		my $sec,  my $min,  my $hour, my $mday, my $mon,
		my $year, my $wday, my $yday, my $isdst
	) = localtime(time);
	$msg = sprintf(
		"<" . $syslogPriTag . ">%s %02d %02d:%02d:%02d",
		$monthMap{ sprintf( "%02d", $mon + 1 ) },
		$mday, $hour, $min, $sec
	);
	$msg = $msg . " " . $syslogHost . " AgentDevice=OracleOSAudit";
	$msg = $msg . "\tSourceFile=" . $currentRec->{SOURCE};
	$msg = $msg . "\tDeviceTime=" . $currentRec->{TIMESTAMP};
	$msg = $msg . "\tDatabaseUser=" . $currentRec->{DB_USER};
	$msg = $msg . "\tPrivilege=" . $currentRec->{PRIVILEGE};
	$msg = $msg . "\tClientUser=" . $currentRec->{CL_USER};
	$msg = $msg . "\tClientTerminal=" . $currentRec->{CL_TERMINAL};
	$msg = $msg . "\tStatus=" . $currentRec->{STATUS};
	$msg = $msg . "\tAction=" . $currentRec->{ACTION};
	# Send the log message.
	$msg = $msg . "\n";

	#Debug
#	print "$msg";

	#Real

	$lastMessage = $msg;
	print "SENDING: $msg\n" if $options{V};
	syswrite( $socket, $msg, length($msg) );
}

sub resetCurrentLogRecord() {

	# Reset our current record.
	my $newRec = {
		TIMESTAMP   => "",
		ACTION      => "",
		DB_USER     => "",
		PRIVILEGE   => "",
		CL_USER     => "",
		CL_TERMINAL => "",
		STATUS      => "",
		SOURCE      => "",
	};

	$currentRec = $newRec;
}

my $interrupted = 0;

# set up signal handlers for proper init control
$SIG{HUP} = sub { 
	print "caught sighup, reset connection\n"; 
	$interrupted = 1;
	$continue = 0; 
}; # don't exit, just reset connection
$SIG{PIPE} = sub { 
	print "caught sigpipe, reset connection\n"; 
	$interrupted = 1;
	$continue = 0;
}; # don't exit, just reset connection
$SIG{INT} = sub {
	print "caught sigint, exiting\n";
	$running  = 0;
	$continue = 0;
	disconnectMe($socket);
};
$SIG{QUIT} = sub {
	print "caught sigquit, exiting\n";
	$running  = 0;
	$continue = 0;
	disconnectMe($socket);
};
$SIG{TERM} = sub {
	print "caught sigterm, exiting\n";
	$running  = 0;
	$continue = 0;
	disconnectMe($socket);
};

if ( !$options{D} ) {
	startDaemon( $execName, $lockFile, $pidFile );
}

my $exit_code = 0;

while ($running) {
	stubbornReconnect( $options{t}, $options{p}, 60 );
	if ( !$continue ) {
		next;
	}

	# now that we're connected, let's see if we can collect some data	
	# Although if we were already running and were interrupted for some reason
 	# we don't want to resend everything we've seen.
	if ( !$interrupted )
	{
		processDirectory( $dirPath );
	}
	else	# Send the event that we lost during interruption
	{
		if ( $lastMessage )
		{	
			# Had some timing issues with setting up the socket and then instantly trying to cram an event through it
			# so we'll wait 1 second just to ensure everything is initialized correctly
			sleep(1);	
			print "About to send message due to disruption: ".$lastMessage."\n";
			print "SENDING: $lastMessage\n" if $options{V};
			syswrite( $socket, $lastMessage, length($lastMessage) );		
			$lastMessage = 0;
		}
		
	}

	$interrupted = 0;
	
	while ( $running && $continue ) {
		sleep(10);

		# Unwatch any files that have been deleted and clear %unwatch files when completed	
		foreach (@unwatchFiles)
		{
			print "Removing deleted file ".$_." from watch list\n";
			$monitor->unwatch($_);
			delete($allLogFiles{$_});
		}
		undef @unwatchFiles;

		#print "Scanning for changes\n";
		$monitor->scan;
	}
}

if ( !$options{D} ) {
	stopDaemon( $lockFile, $pidFile );
}

exit($exit_code);
